<?php
namespace Mewz\WCAS\Actions\Admin\Stock;

use Mewz\Framework\Base\Action;
use Mewz\QueryBuilder\DB;
use Mewz\WCAS\Models\AttributeStock;
use Mewz\WCAS\Util\Matches;

class StockListQuery extends Action
{
	public function __hooks()
	{
		add_action('pre_get_posts', [$this, 'pre_get_posts'], 0);
	}

	public function pre_get_posts(\WP_Query $query)
	{
		if (!$query->is_main_query() || $query->get('post_type') !== 'mewz_attribute_stock') {
			return;
		}

		// remove any stray meta queries added before this point
		// (we need our meta queries to be first to allow sorting by meta_value)
		$query->set('meta_query', []);

		$this->query_orderby($query);
		$this->query_filter_attribute($query);
	}

	public function query_orderby(\WP_Query $query)
	{
		$orderby = $query->get('orderby');
		$_orderby = [];

		if (!$orderby) {
			$orderby = [];
		} elseif (!is_array($orderby)) {
			$orderby = [(string)$orderby => $query->get('order') ?: 'asc'];
		}

		foreach ($orderby as $key => $order) {
			$prop = AttributeStock::get_post_prop($key);

			if (!$prop) {
				$meta_key = '_' . $key;

				$meta_query = [
					'relation' => 'OR',
					['key' => $meta_key, 'compare' => 'NOT EXISTS'],
					['key' => $meta_key],
				];

				$this->add_meta_query($query, $meta_query);

				$prop = $key === 'quantity' ? 'meta_value_num' : 'meta_value';
			}

			if (!isset($_orderby[$prop])) {
				$_orderby[$prop] = $order;
			}
		}

		if (!isset($_orderby['post_title'])) {
			$_orderby['post_title'] = 'asc';
		}

		if (!isset($_orderby['post_date'])) {
			$_orderby['post_date'] = 'desc';
		}

		$query->set('orderby', $_orderby);

		// allow sorting by post_excerpt and post_content
		add_filter('posts_orderby', [$this, 'query_orderby_clause'], 10, 2);
	}

	public function query_orderby_clause($clause, \WP_Query $query)
	{
		global $wpdb;

		if (!$query->is_main_query() || $query->get('post_type') !== 'mewz_attribute_stock') {
			return $clause;
		}

		$orderby = $query->get('orderby');

		if (!$orderby || !is_array($orderby)) {
			return $clause;
		}

		// if the first (main) orderby field is post_excerpt or post_content, add it to the orderby query
		reset($orderby);

		if (in_array($key = key($orderby), ['post_excerpt', 'post_content'])) {
			$clause = "{$wpdb->posts}.{$key} {$orderby[$key]}, {$clause}";
		}

		return $clause;
	}

	public function query_filter_attribute(\WP_Query $query)
	{
		if (!empty($_REQUEST['attribute'])) {
			$filter_term = isset($_REQUEST['term']) && $_REQUEST['term'] !== '';

			$db_query = Matches::query()->where('r.attribute_id', (int)$_REQUEST['attribute']);

			if ($filter_term) {
				$db_query->where('r.term_id', (int)$_REQUEST['term']);
			}

			$stock_ids = $db_query->col('s.stock_id');

			if (!$filter_term) {
				$attr_level_ids = DB::table(DB::$wpdb->posts, 'p')
					->left_join(DB::$wpdb->postmeta, 'pm')->on('pm.post_id = p.ID')
					->where('p.post_type', 'mewz_attribute_stock')
					->where('pm.meta_key', 'attribute_level')
					->where('pm.meta_value', (int)$_REQUEST['attribute'])
					->distinct()
					->col('p.ID');

				if ($attr_level_ids) {
					$stock_ids = array_unique(array_merge($stock_ids, $attr_level_ids));
				}
			}

			if ($stock_ids) {
				$query->set('post__in', $stock_ids);
			} else {
				$query->set('p', -1);
			}
		}
		elseif (!empty($_REQUEST['attribute_level'])) {
			$this->add_meta_query($query, [
				'key' => 'attribute_level',
				'value' => (int)$_REQUEST['attribute_level'],
			]);
		}
	}

	public function add_meta_query(\WP_Query $query, array $meta_query)
	{
	    $meta_queries = $query->get('meta_query') ?: [];
		$meta_queries[] = $meta_query;

		$query->set('meta_query', $meta_queries);
	}
}
