<?php
namespace Mewz\WCAS\Plugin;

use Mewz\Framework\Plugin;
use Mewz\Framework\Util\PostTypes;
use Mewz\WCAS\Actions;
use Mewz\WCAS\Compatibility;

class Loader extends Plugin\Loader
{
	public function load()
	{
		parent::load();

		// load compatibility integration overrides
		add_filter('woocommerce_integrations', [$this, 'integrations_compatibility'], 20);
	}

	public function init()
	{
		// make sure WooCommerce is active before running
		if (!defined('WC_PLUGIN_FILE')) return;

	    parent::init();

		add_action('init', [$this, 'register_post_types'], 5);
		add_action('rest_api_init', [$this, 'load_rest_api']);

		// load common actions
		$this->action(Actions\Common\DeleteMatches::class);
		$this->action(Actions\Common\UpdateOrderStock::class);
		$this->action(Actions\Common\StockNotifications::class);
		$this->action(Actions\Common\ClearCaches::class);

		// load common compatibility actions
		$this->action(Compatibility\Actions\WooCommerce::class);

		if (defined('WPLE_PLUGIN_VERSION') || defined('WPLA_VERSION')) {
			$this->action(Compatibility\Actions\WPLister::class);
		}

		// load admin actions
		if ($this->context->admin) {
			add_action('current_screen', [$this, 'current_screen'], 0);
			add_action('load-product_page_product_attributes', [$this, 'load_product_attributes'], 0);
			add_action('load-woocommerce_page_wc-reports', [$this, 'load_wc_reports'], 0);
			add_action('load-plugins.php', [$this, 'load_plugins_page'], 0);
		}

		// load frontend actions
		if (!MEWZ_WCAS_LITE && apply_filters('mewz_wcas_limit_product_stock', $this->context->front)) {
			$this->action(Actions\Front\ProductLimits::class);

			if ($this->context->front) {
				// load frontend actions
				add_action('wp_loaded', [$this, 'wp_loaded_front'], 0);
				add_action('wc_ajax_get_variation', [$this, 'wc_ajax_get_variation'], 0);

				$this->action(Actions\Front\VariableLimits::class);
			}
		}

		if (!MEWZ_WCAS_LITE && !class_exists(Plugin\Updater::class, false)) {
			$this->plugin->__load_updater();

			if (!class_exists(Plugin\Updater::class, false)) {
				require_once ABSPATH . '/wp-admin/includes/plugin.php';
				deactivate_plugins($this->plugin->basename);
			}
		}
	}

	public function register_post_types()
	{
		PostTypes::add('mewz_attribute_stock', [
			'show_ui' => true,
			'show_in_menu' => 'edit.php?post_type=product',
			'capability_type' => 'product',
			'map_meta_cap' => true,
			'menu_icon' => 'dashicons-clipboard',
			'supports' => ['title'],
			'description' => __('WooCommerce attribute stock items.', 'woocommerce-attribute-stock'),
			'labels' => [
				'name'          => _x('Attribute Stock', 'post type general name', 'woocommerce-attribute-stock'),
				'singular_name' => _x('Attribute Stock', 'post type singular name', 'woocommerce-attribute-stock'),
			],
		]);
	}

	public function register_scripts()
	{
		if ($this->context->admin) {
			$this->scripts->register_css('@admin/attributes');
			$this->scripts->register_js('@admin/attributes', ['translate' => true]);

			$this->scripts->register_css('@admin/stock-list');
			$this->scripts->register_js('@admin/stock-list');

			$this->scripts->register_css('@admin/stock-edit');
			$this->scripts->register_js('@admin/stock-edit', ['translate' => true]);

			$this->scripts->register_css('@admin/reports');
		}
		elseif (!MEWZ_WCAS_LITE) {
			$this->scripts->register_js('@front/variable-stock');
		}
	}

	public function load_rest_api()
	{
		$this->action(Actions\Admin\Stock\StockAjax::class);
	}

	public function admin_init()
	{
		parent::admin_init();

		if ($this->context->taxonomy && strpos($this->context->taxonomy, 'pa_') === 0) {
			$this->action(Actions\Admin\Attributes\AttributeTermEdit::class);
			$this->action(Actions\Admin\Attributes\AttributeTermSave::class);
			$this->action(Actions\Admin\Attributes\AttributeTermList::class);
		}

		$this->action(Actions\Admin\Products\ProductVariationOptions::class);

		$this->action(Actions\Admin\Stock\StockExport::class);

		if (!WP_DEBUG && !MEWZ_WCAS_LITE && !class_exists(Plugin\Authorizer::class, false)) {
			require_once ABSPATH . '/wp-admin/includes/plugin.php';
			deactivate_plugins($this->plugin->basename);
		}
	}

	public function current_screen()
	{
		switch ($this->context->screen_id) {
			case 'edit-mewz_attribute_stock':
				$this->action(Actions\Admin\Stock\StockList::class);
				$this->action(Actions\Admin\Stock\StockListFilters::class);
				$this->action(Actions\Admin\Stock\StockListQuery::class);
				$this->action(Actions\Admin\Stock\StockActions::class);
				$this->action(Actions\Admin\Stock\StockBulkActions::class);
				$this->action(Actions\Admin\PluginHelp::class);
				break;

			case 'mewz_attribute_stock':
				$this->action(Actions\Admin\Stock\StockEdit::class);
				$this->action(Actions\Admin\Stock\StockSave::class);
				$this->action(Actions\Admin\Stock\StockActions::class);
				$this->action(Actions\Admin\PluginHelp::class);
				break;

			case 'product':
				$this->action(Actions\Admin\Products\ProductOptions::class);
				break;
		}
	}

	public function load_product_attributes()
	{
		$this->action(Actions\Admin\Attributes\AttributeEdit::class);
		$this->action(Actions\Admin\Attributes\AttributeSave::class);

		if (!isset($_GET['edit'])) {
			$this->action(Actions\Admin\Attributes\AttributeList::class);
		}
	}

	public function load_wc_reports()
	{
		$this->action(Actions\Admin\Reports\StockReport::class);
	}

	public function load_plugins_page()
	{
		$this->action(Actions\Admin\PluginLinks::class);
	}

	public function wp_loaded_front()
	{
		if (WC()->cart) {
			$this->action(Actions\Front\CartAdd::class);
			$this->action(Actions\Front\CartItems::class);
		}
	}

	public function wc_ajax_get_variation()
	{
		$this->action(Actions\Front\VariableLimitsAjax::class);
	}

	public function integrations_compatibility($integrations)
	{
		// WooCommerce Cart Stock Reducer
		if ($this->context->front && class_exists(\WC_Cart_Stock_Reducer::class, false)) {
			$csr_index = array_search(\WC_Cart_Stock_Reducer::class, $integrations, true);

			if ($csr_index !== false) {
				$integrations[$csr_index] = Compatibility\Components\WCCartStockReducer::class;
			}
		}

		return $integrations;
	}
}
