<?php
namespace Mewz\WCAS\Util;

use Mewz\Framework\Util\WooCommerce;
use Mewz\QueryBuilder\DB;
use Mewz\WCAS\Models\AttributeStock;

class Matches
{
	const SETS_TABLE = 'wc_mewz_wcas_match_sets';
	const ROWS_TABLE = 'wc_mewz_wcas_match_rows';

	/**
	 * @param string $join
	 *
	 * @return \Mewz\QueryBuilder\Query
	 */
	public static function query($join = 'join')
	{
		return DB::table(self::SETS_TABLE, 's')
			->$join(self::ROWS_TABLE, 'r')->on('r.set_id = s.id')
			->distinct();
	}

	/**
	 * Query stock for a specified attribute and optional term. Primarily for admin usage.
	 *
	 * @param int|string $attribute
	 * @param int $term_id
	 * @param string $context 'view' or 'edit'
	 * @param string $return 'object' or 'id'
	 *
	 * @return AttributeStock[]|int[]|\Mewz\QueryBuilder\Query
	 */
	public static function query_stock($attribute, $term_id = null, $context = 'view', $return = 'object')
	{
		$attribute_id = Attributes::get_attribute_id($attribute);

		$cache_key = "query_stock_{$attribute_id}_{$term_id}_{$context}";
		$cache_tags = ['match_sets', 'stock', 'attribute_level'];

		$stock_ids = Mewz_WCAS()->cache->get($cache_key, $cache_tags);

		if ($stock_ids === null) {
			$query = DB::table(DB::$wpdb->posts, 'p')
				->asc('p.post_title')
				->distinct();

			if ($term_id === null) {
				$query->left_join(DB::$wpdb->postmeta, 'pm_al')->on('pm_al.post_id = p.ID AND pm_al.meta_key = ?', 'attribute_level')
					->where('pm_al.meta_value', $attribute_id);
			} else {
				$query->left_join(self::SETS_TABLE, 's')->on('s.stock_id = p.ID')
					->left_join(self::ROWS_TABLE, 'r')->on('r.set_id = s.id')
					->where("s.multiplier != '0'")
					->where('r.attribute_id', $attribute_id)
					->where('r.term_id', $term_id);

				if ($context === 'edit') {
					$query->left_join(DB::$wpdb->postmeta, 'pm_al')->on('pm_al.post_id = p.ID AND pm_al.meta_key = ? AND pm_al.meta_value = ?', 'attribute_level', $attribute_id)
						->is_null('pm_al.meta_value');
				}
			}

			if ($context === 'view') {
				$query->where('p.post_status', 'publish');
			} else {
				$query->not('p.post_status', ['trash', 'auto-draft']);
			}

			$stock_ids = $query->col('p.ID');

			// in view context, only return attribute term level stocks if any exist
			if ($term_id !== null && $context === 'view' && $stock_ids) {
				$attr_level_ids = DB::table(DB::$wpdb->postmeta)
					->where('post_id', $stock_ids)
					->where('meta_key', 'attribute_level')
					->where('meta_value', $attribute_id)
					->col('post_id');

				if (count($attr_level_ids) !== count($stock_ids)) {
					$stock_ids = array_values(array_diff($stock_ids, $attr_level_ids));
				}
			}

			Mewz_WCAS()->cache->set($cache_key, $stock_ids, $cache_tags);
		}

		if (!$stock_ids) return $stock_ids;

		if ($return === 'object') {
			$stocks = [];

			foreach ($stock_ids as $stock_id) {
				$stock = AttributeStock::instance($stock_id, $context);

				if ($stock->exists()) {
					$stocks[] = $stock;
				}
			}

			return $stocks;
		} else {
			return $stock_ids;
		}
	}

	/**
	 * Finds all valid attribute stock items that match a product.
	 *
	 * @param \WC_Product|int $product The product object or ID
	 * @param array $attributes Key/value pairs where key is an attribute id/name/taxonomy
	 *                          and value is a term id/slug or an array of term ids and/or slugs.
	 *                          Passed directly to {@see Matches::match_stock()}.
	 *
	 * @return array
	 */
	public static function match_product_stock($product, array $attributes)
	{
		$matches = [];

		if ($raw_matches = self::match_raw_stock($attributes)) {
			foreach ($raw_matches as $match) {
				$stock = AttributeStock::instance($match['stock_id']);

				if (!$stock->exists() || !$stock->enabled() || !self::validate_filters($stock, $product)) {
					continue;
				}

				if ($stock->match_all()) {
					$match['multiplier'] = $match['multiplier_all'];
				}

				unset($match['multiplier_all']);

				$matches[] = $match;
			}
		}

		/**
		 * IMPORTANT: The results from this filter are cached for performance reasons.
		 * This means the input ($attribute_data) should always have the same output,
		 * or the cache must be invalidated accordingly.
		 *
		 * @see Limits::get_stock_limits()
		 */
		return apply_filters('mewz_wcas_product_stock_matches', $matches, $product, $attributes);
	}

	/**
	 * Matches a set of attributes/terms to stock items.
	 *
	 * @param array $attributes Key/value pairs where key is an attribute id/name/taxonomy
	 *                          and value is a term id/slug or an array of term ids and/or slugs
	 *
	 * @return array Raw stock match results, before any validation or filtering
	 */
	public static function match_raw_stock(array $attributes)
	{
		global $wpdb;

		if (!$attributes) return [];

		$attributes = Attributes::get_attribute_id_sets($attributes);

		$cache_key = 'match_stock_' . md5(json_encode($attributes));
		$cache_tags = ['match_sets'];

		$matches = Mewz_WCAS()->cache->get($cache_key, $cache_tags);

		if ($matches === null) {
			$conditions = [];

			foreach ($attributes as $attribute_id => $term_ids) {
				foreach ($term_ids as $term_id) {
					$condition = 'r.attribute_id = ' . $attribute_id . ' AND r.term_id ';
					$condition .= $term_id ? 'IN (' . $term_id . ', 0)' : '= 0';

					$conditions[] = "($condition)";
				}
			}

			if (!$conditions) return [];

			$conditions = implode("\nOR ", $conditions);

			$rows_table = $wpdb->prefix . self::ROWS_TABLE;
			$sets_table = $wpdb->prefix . self::SETS_TABLE;

			// not using query builder here to save a little bit on performance as it's a fairly complex query
			// and this method will often be called many times in a row (e.g. when updating several orders at once)
			$query = "
				SELECT stock_id, multiplier, SUM(multiplier) AS multiplier_all
				FROM (
					SELECT s.stock_id, r.set_id, IF(s.multiplier = '', 1, +s.multiplier) AS multiplier
					FROM {$sets_table} s
					JOIN {$rows_table} r ON r.set_id = s.id
					WHERE s.multiplier != '0' AND {$conditions}
					GROUP BY r.set_id
					HAVING COUNT(r.id) = (
						SELECT COUNT(*) FROM {$rows_table} r2
						WHERE r2.set_id = r.set_id
					)
					ORDER BY s.priority
				) matches
				GROUP BY stock_id
			";

			$matches = $wpdb->get_results($query, ARRAY_A);

			if ($matches) {
				foreach ($matches as &$match) {
					$match['stock_id'] = (int)$match['stock_id'];
					$match['multiplier'] = (float)$match['multiplier'];
					$match['multiplier_all'] = (float)$match['multiplier_all'];
				}
			}

			Mewz_WCAS()->cache->set($cache_key, $matches, $cache_tags);
		}

		return $matches ?: [];
	}

	/**
	 * Check all configured attribute stock item filters to see if product should be matched.
	 *
	 * @param AttributeStock $stock
	 * @param \WC_Product|\WC_Product[] $product
	 *
	 * @return bool|mixed
	 */
	public static function validate_filters(AttributeStock $stock, $product)
	{
		if (is_array($product)) {
			foreach ($product as $prod) {
				$valid = self::_validate_filters($stock, $prod);
				$valid = apply_filters('mewz_wcas_validate_stock_filters', $valid, $stock, $prod);

				if ($valid) return true;
			}
		} else {
			$valid = self::_validate_filters($stock, $product);
			return apply_filters('mewz_wcas_validate_stock_filters', $valid, $stock, $product);
		}
	}

	protected static function _validate_filters(AttributeStock $stock, \WC_Product $product)
	{
		$product_id = $product->get_id();
		$parent_id = $product->get_parent_id();

		// if product or product's parent is specifically excluded, immediately fail
		$excl_products = $stock->exclude_products();

		if ($excl_products && (in_array($product_id, $excl_products) || ($parent_id && in_array($parent_id, $excl_products)))) {
			return false;
		}

		// if product or product's parent is specifically included, immediately pass
		$incl_products = $stock->products();

		if ($incl_products && (in_array($product_id, $incl_products) || ($parent_id && in_array($parent_id, $incl_products)))) {
			return true;
		}

		// if product type isn't valid, fail
		if ($valid_types = $stock->product_types()) {
			$product_type = \WC_Product_Factory::get_product_type($parent_id ?: $product_id);

			if (!in_array($product_type, $valid_types)) {
				return false;
			}
		}

		// check category filters
		$incl_cats = $stock->categories();
		$excl_cats = $stock->exclude_categories();

		if ($incl_cats || $excl_cats) {
			$product_cats = Products::get_all_product_category_ids($product);

			// if none of the product's categories are descendants of the included categories, fail
			if ($incl_cats && (!$product_cats || !array_intersect($product_cats, $incl_cats))) {
				return false;
			}

			// if any of the product's categories are descendants of the excluded categories, fail
			if ($excl_cats && $product_cats && array_intersect($product_cats, $excl_cats)) {
				return false;
			}
		}

		// if product filters are specified but don't match, and no other filters are specified, fail
		if ($incl_products && !$valid_types && !$incl_cats && !$excl_cats) {
			return false;
		}

		return true;
	}

	/**
	 * @param int $stock_id
	 * @param string $context 'view' or 'edit'
	 *
	 * @return array [set_id] => [multiplier, rows[] => [attribute, term]]
	 */
	public static function get_sets($stock_id, $context = 'view')
	{
		$cache_key = "match_sets_{$stock_id}_{$context}_" . WooCommerce::get_cache_incr('woocommerce-attributes');
		$sets = Mewz_WCAS()->cache->get($cache_key, 'match_sets');

		if ($sets === null) {
			$sets = [];

			$results = self::query()
				->select('r.set_id', 'r.id row_id', 's.multiplier', 'r.attribute_id', 'r.term_id')
				->asc('s.priority')
				->find_all('s.stock_id', $stock_id);

			if ($results) {
				$attributes = Attributes::get_attributes();

				foreach ($results as $result) {
					$set_id = (int)$result['set_id'];

					if (!isset($sets[$set_id])) {
						$multiplier = $result['multiplier'];

						if ($context === 'view') {
							$multiplier = $multiplier === '' ? 1 : (float)$multiplier;
						}

						$sets[$set_id] = [
							'multiplier' => $multiplier,
							'rows' => [],
						];
					}

					$attribute_id = (int)$result['attribute_id'];

					if (isset($attributes[$attribute_id])) {
						$sets[$set_id]['rows'][$attribute_id] = [
							'attribute' => $attribute_id,
							'term' => (int)$result['term_id'],
						];
					}
				}

				$row_sort_callback = static function($row1, $row2) use ($attributes) {
					return strnatcmp(
						$attributes[$row1['attribute']]['label'],
						$attributes[$row2['attribute']]['label']
					);
				};

				foreach ($sets as $set_id => &$set) {
					$row_count = count($set['rows']);

					if ($row_count === 0) {
						unset($sets[$set_id]);
					} elseif ($row_count > 1) {
						usort($set['rows'], $row_sort_callback);
					} else {
						$set['rows'] = array_values($set['rows']);
					}
				}
			}

			Mewz_WCAS()->cache->set($cache_key, $sets, 'match_sets');
		}

		if ($context === 'view') {
			$sets = apply_filters('mewz_wcas_get_match_sets', $sets, $stock_id);
		}

		return $sets;
	}

	/**
	 * @param int $stock_id
	 * @param array $sets Set format: [multiplier, priority, rows[] => [attribute, term]]
	 */
	public static function save_sets($stock_id, $sets)
	{
		self::delete_sets($stock_id, null, false);

		$priority = 0;

		foreach ($sets as $set) {
			// make sure the set has at least one non-empty row
			if (!$set['rows'] || !array_filter(array_column($set['rows'], 'attribute'))) {
				continue;
			}

			if (isset($set['multiplier']) && $set['multiplier'] !== '') {
				$multiplier = is_string($set['multiplier']) ? (float)trim($set['multiplier']) : $set['multiplier'];
			} else {
				$multiplier = '';
			}

			$priority = isset($set['priority']) ? (int)$set['priority'] : $priority + 1;

			$set_id = DB::insert(self::SETS_TABLE, [
				'stock_id' => $stock_id,
				'multiplier' => $multiplier,
				'priority' => $priority,
			]);

			$attribute_rows = [];

			foreach ($set['rows'] as $row) {
				if (!empty($row['attribute']) && isset($row['term']) && $row['term'] !== '' && (int)$row['term'] >= 0) {
					$attribute_rows[] = [
						'set_id' => $set_id,
						'attribute_id' => (int)$row['attribute'],
						'term_id' => (int)$row['term'],
					];
				}
			}

			DB::insert(self::ROWS_TABLE, $attribute_rows);
		}

		Mewz_WCAS()->cache->invalidate('match_sets');

		do_action('mewz_wcas_match_sets_saved', $stock_id, $sets);
	}

	/**
	 * @param int $stock_id
	 * @param int|string $attribute
	 * @param int $term_id
	 * @param string|float $multiplier
	 *
	 * @return array [set_id, row_id]
	 */
	public static function add_set($stock_id, $attribute, $term_id = 0, $multiplier = '')
	{
		$attribute_id = Attributes::get_attribute_id($attribute);

		$max_priority = DB::table(self::SETS_TABLE)
			->where('stock_id', $stock_id)
			->val('MAX(priority) priority');

		$priority = (int)$max_priority + 1;
		$multiplier = (string)$multiplier === '' ? (string)$multiplier : (string)(float)$multiplier;

		$set_id = DB::insert(self::SETS_TABLE, [
			'stock_id' => $stock_id,
			'multiplier' => $multiplier,
			'priority' => $priority,
		]);

		$row_id = DB::insert(self::ROWS_TABLE, [
			'set_id' => $set_id,
			'attribute_id' => $attribute_id,
			'term_id' => $term_id,
		]);

		Mewz_WCAS()->cache->invalidate('match_sets');

		return compact('set_id', 'row_id');
	}

	/**
	 * @param int $stock_id
	 * @param int $set_id
	 * @param bool $invalidate_cache
	 */
	public static function delete_sets($stock_id, $set_id = null, $invalidate_cache = true)
	{
		$query = self::query('left_join')
			->table('r', true)
			->where('s.stock_id', $stock_id);

		if ($set_id !== null) {
			$query->where('s.id', $set_id);
		}

		$query->delete();

		// reset id auto increments to next highest values
		DB::table(self::SETS_TABLE)->alter('AUTO_INCREMENT = 1');
		DB::table(self::ROWS_TABLE)->alter('AUTO_INCREMENT = 1');

		if ($invalidate_cache) {
			Mewz_WCAS()->cache->invalidate('match_sets');
		}

		do_action('mewz_wcas_match_sets_deleted', $stock_id, $set_id);
	}

	public static function delete_rows($attribute, $term_id = null, $trash_attribute_level = false)
	{
		$attribute_id = Attributes::get_attribute_id($attribute);

		// get affected stock ids
		$query = self::query()->where('r.attribute_id', $attribute_id);

		if ($term_id !== null) {
			$query->where('r.term_id', $term_id);
		}

		$stock_ids = $query->col('s.stock_id');
		if (!$stock_ids) return;

		// delete all matching attribute match-set rows
		$query = DB::table(self::ROWS_TABLE)->where('attribute_id', $attribute_id);

		if ($term_id !== null) {
			$query->where('term_id', $term_id);
		}

		$deleted_rows = $query->delete();
		if (!$deleted_rows) return;

		DB::table(self::ROWS_TABLE)->alter('AUTO_INCREMENT = 1');

		// invalidate all caches with the 'match_sets' tag
		Mewz_WCAS()->cache->invalidate('match_sets');

		// delete empty match sets remaining
		$query = self::query('left_join')
			->where('s.stock_id', $stock_ids)
			->is_null('r.id');

		$stock_ids = $query->col('s.stock_id');
		if (!$stock_ids) return;

		$query->delete();

		DB::table(self::SETS_TABLE)->alter('AUTO_INCREMENT = 1');

		// handle empty stocks remaining
		$remaining_stock_ids = DB::table(self::SETS_TABLE)
			->where('stock_id', $stock_ids)
			->distinct()
			->col('stock_id');

		if ($remaining_stock_ids) {
			$empty_stock_ids = array_diff($stock_ids, $remaining_stock_ids);
		} else {
			$empty_stock_ids = $stock_ids;
		}

		if ($empty_stock_ids) {
			foreach ($empty_stock_ids as $stock_id) {
				if (!$trash_attribute_level && $attribute_level_id = get_post_meta($stock_id, 'attribute_level', true)) {
					self::add_set($stock_id, $attribute_level_id);
				} else {
					(new AttributeStock($stock_id, 'object'))->trash();
				}
			}
		}
	}

	public static function delete_empty_attribute_level_stocks($attribute, $permanent = false)
	{
		$attribute_id = Attributes::get_attribute_id($attribute);

		$args = [
			'meta_key' => 'attribute_level',
			'meta_value' => $attribute_id,
		];

		$stock_ids = AttributeStock::query($args, 'edit', 'id');

		if ($stock_ids) {
			foreach ($stock_ids as $stock_id) {
				if (count(get_post_meta($stock_id, 'attribute_level')) === 1 && !self::get_sets($stock_id, 'edit')) {
					(new AttributeStock($stock_id, 'object'))->delete($permanent);
				}
			}
		}

		AttributeStock::delete_all_meta('attribute_level', $attribute_id);

		Mewz_WCAS()->cache->invalidate('attribute_level');
	}

	public static function get_stock_attributes($post_status = 'all')
	{
		if ($post_status === 'all') {
			$post_status = ['publish', 'draft'];
		}

		$rows = self::query()
			->join(DB::$wpdb->posts, 'p')->on('p.ID = s.stock_id')
			->left_join(DB::$wpdb->postmeta, 'pm_al')->on('pm_al.post_id = p.ID AND pm_al.meta_key = ?', 'attribute_level')
			->where('p.post_status', $post_status)
			->select('r.attribute_id', 'r.term_id')
			->distinct()
			->get();

		$attr_level_ids = DB::table(DB::$wpdb->postmeta, 'pm')
			->join(DB::$wpdb->posts, 'p')->on('p.ID = pm.post_id')
			->where('p.post_type', 'mewz_attribute_stock')
			->where('p.post_status', $post_status)
			->where('pm.meta_key', 'attribute_level')
			->distinct()
			->col('pm.meta_value');

		$attributes = [];

		if ($rows) {
			foreach ($rows as $row) {
				$attributes[$row['attribute_id']][$row['term_id']] = (int)$row['term_id'];
			}
		}

		if ($attr_level_ids) {
			foreach ($attr_level_ids as $attr_id) {
				if (!isset($attributes[$attr_id])) {
					$attributes[$attr_id] = [];
				}
			}
		}

		return $attributes;
	}

	/**
	 * @param int $stock_id
	 * @param int|string $attribute
	 * @param int $term_id
	 *
	 * @return float
	 */
	public static function get_term_quantity_multiplier($stock_id, $attribute, $term_id)
	{
		$attribute_id = Attributes::get_attribute_id($attribute);

		$multiplier = self::query()
			->where('s.stock_id', $stock_id)
			->where('r.attribute_id', $attribute_id)
			->where('r.term_id', $term_id)
			->asc('s.priority')
			->val('s.multiplier');

		return $multiplier === null ? 1 : (float)$multiplier;
	}

	/**
	 * @param AttributeStock|int $stock
	 * @param int|string $attribute
	 * @param int $term_id
	 *
	 * @return float
	 */
	public static function get_term_available_quantity($stock, $attribute, $term_id)
	{
		if (!$stock instanceof AttributeStock) {
			$stock = AttributeStock::instance($stock);
		}

		$multiplier = self::get_term_quantity_multiplier($stock->id(), $attribute, $term_id);

		if ($multiplier == 1) {
			return $stock->quantity();
		} else {
			return floor($stock->quantity() / $multiplier);
		}
	}

	/**
	 * @param AttributeStock|int $stock
	 * @param int|string $attribute
	 * @param int $term_id
	 *
	 * @return float
	 */
	public static function get_term_display_quantity($stock, $attribute, $term_id)
	{
		$available = self::get_term_available_quantity($stock, $attribute, $term_id);
		$quantity = $stock->quantity();

		if ($available != $stock->quantity()) {
			$available .= ' (' . $quantity . ')';
		}

		return $available;
	}

	/**
	 * @param int[] $stock_ids
	 * @param int|string $attribute_id
	 * @param int $term_id
	 *
	 * @return string
	 */
	public static function get_quantity_range($stock_ids, $attribute_id, $term_id = null)
	{
		$quantities = [];

		foreach ($stock_ids as $stock_id) {
			$stock = AttributeStock::instance($stock_id);

			// show available stock for terms of attribute-level stock when displaying a quantity range
			if ($term_id && in_array($attribute_id, $stock->meta('attribute_level', false))) {
				$quantities[] = self::get_term_available_quantity($stock_id, $attribute_id, $term_id);
			} else {
				$quantities[] = $stock->quantity();
			}
		}

		$min_qty = min($quantities);
		$max_qty = max($quantities);

		if ($min_qty == $max_qty) {
			return $min_qty;
		} else {
			return $min_qty . ' ... ' . $max_qty;
		}
	}

	/**
	 * @param \WC_Product|\WC_Product[] $product
	 * @param array $attributes
	 * @param bool $include_max
	 *
	 * @return array
	 */
	public static function get_all_match_data($product, array $attributes, $include_max = false)
	{
		if (!$attributes) return [];

		$attribute_ids = Attributes::get_attribute_id_sets($attributes);

		if (!$attribute_ids) return [];

		$set_ids = self::query()
			->join(DB::$wpdb->posts, 'p')->on('p.ID = s.stock_id')
			->where('p.post_type', AttributeStock::POST_TYPE)
			->where('p.post_status', 'publish')
			->where("s.multiplier != '0'")
			->where('r.attribute_id', array_keys($attribute_ids))
			->where('r.term_id', array_merge([0], ...$attribute_ids))
			->col('s.id');

		if (!$set_ids) return [];

		// get ALL match rows from found stock match sets
		$rows = self::query()
			->select('s.stock_id', 'r.set_id', 's.multiplier', 'r.attribute_id', 'r.term_id')
			->where('s.id', $set_ids)
			->asc('s.stock_id', 's.priority')
			->get();

		if (!$rows) return [];

		$matches = [];
		$attr_cache = [];
		$max_quantity = $include_max ? 0 : null;

		foreach ($rows as $row) {
			$stock_id = (int)$row['stock_id'];

			if (isset($skip_stock_ids[$stock_id])) continue;

			$stock = AttributeStock::instance($stock_id);

			if (!$stock->exists() || !$stock->enabled() || !$stock->limit_products() || !self::validate_filters($stock, $product)) {
				$skip_stock_ids[$stock_id] = true;
				continue;
			}

			$set_id = (int)$row['set_id'];
			$attr_id = (int)$row['attribute_id'];

			if (isset($attr_cache[$attr_id])) {
				$attr_name = $attr_cache[$attr_id];
			} else {
				$attr_name = $attr_cache[$attr_id] = Attributes::get_attribute_name((int)$row['attribute_id']);
			}

			if ($term_id = (int)$row['term_id']) {
				$term_slug = Attributes::get_term_prop($term_id, 'pa_' . $attr_name, 'slug');
				if ($term_slug === false) continue;
			} else {
				$term_slug = ''; // any
			}

			if (!isset($matches[$stock_id])) {
				$quantity = $stock->quantity();
				$matches[$stock_id]['q'] = $quantity;

				if ($include_max && $quantity > $max_quantity) {
					$max_quantity = $quantity;
				}

				if ($stock->match_all()) {
					$matches[$stock_id]['ma'] = true;
				}
			}

			if (!isset($matches[$stock_id]['s'][$set_id])) {
				$multiplier = $row['multiplier'] ? (float)$row['multiplier'] : 1;

				if ($multiplier != 1) {
					$matches[$stock_id]['s'][$set_id]['x'] = $multiplier;
				}
			}

			$matches[$stock_id]['s'][$set_id]['r'][$attr_name] = $term_slug;
		}

		$matches = apply_filters('mewz_wcas_all_product_match_data', $matches, $product, $attributes, $max_quantity);

		if ($matches) {
			// remove id keys from lists as they're not needed
			foreach ($matches as $stock_id => $match) {
				$matches[$stock_id]['s'] = array_values($match['s']);
			}

			$matches = array_values($matches);
		}

		if ($include_max) {
			$matches['max_quantity'] = $max_quantity;
		}

		return $matches;
	}

	/**
	 * Finds all match sets containing attributes/terms.
	 *
	 * @param array $attributes
	 * @param bool $exists
	 *
	 * @return array|bool
	 */
	public static function get_all_match_sets(array $attributes, $exists = false)
	{
		if (!$attributes) return [];

		$attribute_ids = Attributes::get_attribute_id_sets($attributes);

		if (!$attribute_ids) return [];

		$query = self::query()
			->join(DB::$wpdb->posts, 'p')->on('p.ID = s.stock_id')
			->where('p.post_type', AttributeStock::POST_TYPE)
			->where('p.post_status', 'publish')
			->where("s.multiplier != '0'")
			->where('r.attribute_id', array_keys($attribute_ids))
			->where('r.term_id', array_merge([0], ...$attribute_ids));

		if ($exists) {
			return $query->exists();
		} else {
			return $query->col('s.id');
		}
	}

	/**
	 * @param \WC_Product|int $product
	 * @param bool $check_parent
	 *
	 * @return bool
	 */
	public static function product_is_limited_by_attribute_stock($product, $check_parent = true)
	{
		static $result = [];

		if (!$product) return false;

		if ($is_object = ($product instanceof \WC_Product)) {
			$product_id = $product->get_id();
		} else {
			$product_id = $product;
		}

		if (isset($result[$product_id])) {
			return $result[$product_id];
		}

		if (!$is_object) {
			$product = wc_get_product($product_id);
			if (!$product) return false;
		}

		if ($check_parent) {
			$parent_id = $product->get_parent_id();

			if ($parent_id > 0) {


				return self::product_matches_any_attribute_stock($parent_id, false);
			}
		}

		$attributes = Products::get_product_attributes($product, null, true);
	}
}
